/*
 * KDE Radio Station
 * Copyright (C) 2003 Josef Spillner <spillner@kde.org>
 * Published under GNU GPL conditions.
 */

#include "stationlist.h"
#include "options.h"
#include "share.h"

#include <kpushbutton.h>
#include <kstdguiitem.h>
#include <klocale.h>
#include <klistview.h>
#include <kio/netaccess.h>
#include <kprocess.h>
#include <kmessagebox.h>
#include <ktempfile.h>
#include <kdebug.h>
#include <kstandarddirs.h>
#include <kconfig.h>
#include <ksimpleconfig.h>
#include <kapplication.h>
#include <kdeversion.h>
#include <kurldrag.h>
#include <kmediaplayer/player.h>
#include <dcopref.h>
#include <dcopclient.h>

#include <qlayout.h>
#include <qdom.h>
#include <qpixmap.h>
#include <qlistview.h>
#include <qfile.h>
#include <qregexp.h>
#include <qtimer.h>
#include <qtextstream.h>

#include <ctime>

#ifdef KDE_IS_VERSION
#if KDE_IS_VERSION(3, 1, 92)
#define NETACCESS_HACK ,this
#else
#define NETACCESS_HACK
#endif
#else
#define NETACCESS_HACK
#endif

StationList::StationList(QWidget *parent, const char *name)
: KListView(parent, name)
{
	QVBoxLayout *vbox;
	KConfig *config;
	QString sync, cache;
	int lastupdate;
	time_t time;

	view = new KListView(this);
	view->setAllColumnsShowFocus(true);
	view->addColumn(i18n("Stream"));
	view->addColumn(i18n("Bandwidth"));
	view->addColumn(i18n("Style"));
	view->addColumn(i18n("Location"));
	view->addColumn(i18n("URI"));
	view->addColumn(i18n("Type"));

	vbox = new QVBoxLayout(this, 5);
	vbox->add(view);

	m_query = "<?xml version=\"1.0\"><query class=\"metasound\" type=\"connection\" delta=\"%1\">0.1</query>\n";
	m_metaquery = "<?xml version=\"1.0\"><query class=\"metasound\" type=\"meta\">0.1</query>\n";

	m_update = "<?xml version=\"1.0\"?>";
	m_update += "<update class=\"metasound\" type=\"connection\" username=\"%1\" password=\"%2\">";
	m_update += "<option name=\"mode\">add</option>";
	m_update += "<option name=\"version\">0.1</option>";
	m_update += "<option name=\"stream\">%3</option>";
	m_update += "<option name=\"uri\">%4</option>";
	m_update += "<option name=\"location\">%5</option>";
	m_update += "<option name=\"speed\">%6</option>";
	m_update += "<option name=\"style\">%7</option>";
	m_update += "<option name=\"type\">%8</option>";
	m_update += "</update>\n";

	connect(view, SIGNAL(executed(QListViewItem *)), SLOT(slotActivate(QListViewItem *)));

	config = kapp->config();
	config->setGroup("meta");
	m_metaservers = config->readListEntry("metaservers");

	config->setGroup("settings");
	sync = config->readEntry("synchronization", "weekly");
	cache = config->readEntry("cache", "yes");

	show();

	if(cache == "yes")
		loadcache();

	if(sync == "startup")
		slotUpdateStations();
	else if(sync == "weekly")
	{
		lastupdate = config->readNumEntry("lastupdate");
		time = ::time(NULL);
		if(time - lastupdate > 7 * 86400)
		{
			slotUpdateStations();
			config->writeEntry("lastupdate", time);
			config->sync();
		}
	}

	view->setDragEnabled(true);
}

StationList::~StationList()
{
}

void StationList::startDrag()
{
	QListViewItem *item = view->currentItem();

	if(item)
	{
		KURLDrag *d = new KURLDrag(KURL::List(KURL(item->text(4))), this, "DragObject");
		d->dragCopy();
	}
}

QStringList StationList::metaservers(bool writeable)
{
	QStringList tmp;
	QStringList::iterator it;
	KConfig *config;
	bool w;

	if(writeable)
	{
		config = kapp->config();
		config->setGroup("contributions");
		for(it = m_metaservers.begin(); it != m_metaservers.end(); it++)
		{
			w = config->readBoolEntry((*it));
			if(w) tmp << (*it);
		}
		return tmp;
	}
	else return m_metaservers;
}

void StationList::addStation(QString metaserver, QString stream, QString uri, QString location,
	QString speed, QString style, QString type)
{
	QString username, password;
	QString metauri;
	QString update;
	QStringList::iterator it;
	KConfig *config;
	QString cleanurl;

	username = "";
	password = "";

	update = m_update.arg(username
		).arg(password
		).arg(stream
		).arg(uri
		).arg(location
		).arg(speed
		).arg(style
		).arg(type);
	update.replace("&", "&amp;");
	kdDebug() << "UPDATE: " << update << endl;

	config = kapp->config();
	config->setGroup("identifiers");

	metauri = metaserver;
	for(it = m_metaservers.begin(); it != m_metaservers.end(); it++)
	{
		cleanurl = (*it);
		if(config->readEntry(cleanurl.replace("=", "%3d")) == metaserver) metauri = (*it);
	}

	doupdate(update, metauri);
}

void StationList::slotUpdateMetaservers()
{
	doconnection(m_metaquery);
}

void StationList::slotUpdateStations()
{
	doconnection(m_query);
}

void StationList::doconnection(QString query)
{
	QStringList::iterator it;
	bool ret;
	int size;
	QString tmpbuf, tmpname;
	QStringList tmpservers;
	KConfig *config;
	QString alias, cleanurl;
	QTimer timer;
	time_t time;
	int oldtime;

	connect(&timer, SIGNAL(timeout()), SLOT(slotTimeout()));

	m_curquery = query;

	config = kapp->config();

	for(QStringList::iterator it = m_metaservers.begin(); it != m_metaservers.end(); it++)
	{
		config->setGroup("identifiers");
		cleanurl = (*it);
		alias = config->readEntry(cleanurl.replace("=", "%3d"));
		if(alias.isNull()) alias = (*it);
		config->setGroup("meta");
		if(config->readNumEntry(alias) != 1) tmpservers += (*it);
	}

//	if(query != m_metaquery)
//		view->clear(); // FIXME: add/remove protocol

	for(it = tmpservers.begin(); it != tmpservers.end(); it++)
	{
		QString buf;
		QString tmp = (*it);
		kdDebug() << "Synchronize resource: " << tmp << endl;

		if(tmp.startsWith("ggzmeta://"))
		{
			KURL url(tmp);
			m_host = url.host();
			m_port = url.port();
			if(!m_port) m_port = 15689;

			config = kapp->config();
			config->setGroup("deltas");
			oldtime = config->readNumEntry(tmp);
			if(query == m_query)
				m_curquery = query.arg(oldtime);
			else
				m_curquery = query;
			time = ::time(NULL);
			config->writeEntry(tmp, time);
			config->sync();

			m_synchronized = 0;

			sock = new QSocket();
			connect(sock, SIGNAL(connected()), SLOT(slotConnected()));
			connect(sock, SIGNAL(readyRead()), SLOT(slotRead()));
			connect(sock, SIGNAL(error(int)), SLOT(slotError(int)));
			sock->connectToHost(m_host, m_port);

			timer.start(10000, true);
			while(!m_synchronized)
				kapp->processEvents();
			timer.stop();
		}
		else if(query != m_metaquery)
		{
			KURL uri(tmp);
			ret = KIO::NetAccess::download(uri, tmpname NETACCESS_HACK);
			if(ret)
			{
				QFile f(tmpname);
				f.open(IO_ReadOnly);
				while((size = f.readLine(tmpbuf, 1024)) != -1)
				{
					buf += tmpbuf;
				}
				f.close();

				QRegExp exp("\\<\\!\\-\\-([^>])*\\>");
				buf.replace(exp, "");
				buf = QString::fromUtf8(buf);

				QRegExp icecast("\\<entry_list\\>");
				QRegExp ggzmeta("\\<resultset referer=\"([^\"])*\"\\>");
				QRegExp metaservlocal("\\<metaserver\\>");

				if(icecast.search(buf, 0, QRegExp::CaretAtZero) >= 0)
				{
					kdDebug() << "## detected format: icascast" << endl;
					processicecast(buf);
				}
				else if(ggzmeta.search(buf, 0, QRegExp::CaretAtZero) >= 0)
				{
					kdDebug() << "## detected format: ggzmeta" << endl;
					kdDebug() << "## shouldn't happen, not handled yet" << endl;
				}
				else if(metaservlocal.search(buf, 0, QRegExp::CaretAtZero) >= 0)
				{
					kdDebug() << "## detected format: metaservlocal" << endl;
					processlocal(buf);
				}
				else kdDebug() << "## format not recognized" << endl;
			}
		}
	}
}

void StationList::doupdate(QString update, QString uri)
{
	m_curquery = update;

	KURL url(uri);

	sock = new QSocket();
	connect(sock, SIGNAL(connected()), SLOT(slotConnected()));
	connect(sock, SIGNAL(readyRead()), SLOT(slotRead()));
//	connect(sock, SIGNAL(error(int)), SLOT(slotError(int)));
	sock->connectToHost(url.host(), url.port());
}

void StationList::slotConnected()
{
kdDebug() << "QUERY:" << m_curquery << endl;
	sock->writeBlock(m_curquery, m_curquery.length());
	sock->flush();
}

void StationList::slotError(int error)
{
	kdDebug() << "error!" << endl;
	m_synchronized = 1;
}

void StationList::processicecast(QString content)
{
	QDomDocument dom;
	QDomNode node;
	QDomElement element, child, streamchild;
	QString buf, tmp, value, key;

	dom.setContent(content);
	node = dom.documentElement().firstChild();

	if(!node.isNull())
	{
		element = dom.documentElement();
	}

	buf = "<resultset referer=\"\">";

	while(!node.isNull())
	{
		element = node.toElement();
		if((element.tagName() == "entry") && (!element.attribute("name").isNull()))
		{
			child = element.firstChild().toElement();
			while(!child.isNull())
			{
				if(child.tagName() == "stream")
				{
					streamchild = child.firstChild().toElement();

					tmp = "<result preference=\"0\">";
					while(!streamchild.isNull())
					{
						key = streamchild.tagName();
						value = streamchild.text();
						if(key == "listen_url")
						{
							tmp += "<uri>" + value + "</uri>";
						}
						else if(key == "stream_type")
						{
							if(value == "Ogg Vorbis")
								tmp += "<type>stream</type>";
							else if(value == "MP3 audio")
								tmp += "<type>stream</type>";
							else
								tmp += "<type>unknown</type>";
						}
						else if(key == "stream_description")
						{
							tmp += "<stream>" + value + "</stream>";
						}
						else if(key == "current_song")
						{
							// ignore
						}
						else if(key == "genre")
						{
							tmp += "<style>" + value + "</style>";
						}
						else if(key == "audio_info")
						{
							tmp += "<speed>" + value + "</speed>";
						}
						streamchild = streamchild.nextSibling().toElement();
					}
					tmp += "</result>";

					buf += tmp;
				}

				child = child.nextSibling().toElement();
			}
		}
		node = node.nextSibling();
	}

	buf += "</resultset>";

kdDebug() << "processicecast delegates to process(" << buf << ")" << endl;
	process(buf);
}

void StationList::processlocal(QString content)
{
	QDomDocument dom;
	QDomNode node;
	QDomElement element, child;
	QString buf, tmp, value;
	QDomNamedNodeMap map;

	dom.setContent(content);
	node = dom.documentElement().firstChild();

	if(!node.isNull())
	{
		element = dom.documentElement();
	}

	buf = "<resultset referer=\"\">";

	while(!node.isNull())
	{
		element = node.toElement();
		if((element.tagName() == "entrylist") && (element.attribute("class") == "metasound"))
		{
			child = element.firstChild().toElement();
			while(!child.isNull())
			{
				if(child.tagName() == "connection")
				{
					tmp = "<result preference=\"0\">";
					value = child.text();
					value.replace("&amp;", "&");
					value.replace("&", "&amp;");
					tmp += "<uri>" + value + "</uri>";
					map = child.attributes();
					for(unsigned int i = 0; i < map.count(); i++)
					{
						if(map.item(i).nodeName() != "version")
						{
							tmp += "<" + map.item(i).nodeName() + ">";
							tmp += map.item(i).nodeValue();
							tmp += "</" + map.item(i).nodeName() + ">";
						}
					}
					tmp += "</result>";

					buf += tmp;
				}

				child = child.nextSibling().toElement();
			}
		}
		node = node.nextSibling();
	}

	buf += "</resultset>";

kdDebug() << "processlocal delegates to process(" << buf << ")" << endl;
	process(buf);
}

void StationList::process(QString content)
{
	QDomDocument dom;
	QDomNode node;
	QDomElement element, child;
	QString style, stream, speed, uri, location, type;
	QString name;
	QString grouping, cache;
	KListViewItem *tmp;
	QListViewItem *qtmp, *qtmp2;
	KStandardDirs d;
	KConfig *config;
	QString response;
	QStringList::iterator it;
	bool has_metaserver = false;
	bool has_stations = false;
	bool consideration;
	bool writeable;
	QString cleanurl;

	config = kapp->config();
	config->setGroup("settings");
	grouping = config->readEntry("grouping", "flat");
	cache = config->readEntry("cache", "yes");

	if(grouping == "flat")
		view->setRootIsDecorated(false);
	else
		view->setRootIsDecorated(true);

	dom.setContent(content);
	node = dom.documentElement().firstChild();

	if(!node.isNull())
	{
		element = dom.documentElement();
		kdDebug() << "we're at " << element.tagName() << endl;
		if(element.attribute("referer") == "update")
		{
			child = element.firstChild().toElement();
			response = child.text();
			KMessageBox::information(this, i18n("Server response was: %1").arg(response), i18n("Update"));
			return;
		}
	}

	while(!node.isNull())
	{
		stream = "???";
		speed = "???";
		style = "???";
		uri = "???";
		location = "???";
		type = "???";

		name = QString::null;
		writeable = false;

		element = node.toElement();
		child = element.firstChild().toElement();
		while(!child.isNull())
		{
			if(child.tagName() == "stream") stream = child.text();
			else if(child.tagName() == "speed") speed = child.text();
			else if(child.tagName() == "style") style = child.text();
			else if(child.tagName() == "uri") uri = child.text();
			else if(child.tagName() == "location") location = child.text();
			else if(child.tagName() == "type") type = child.text();
			else if(child.tagName() == "name") name = child.text();
			else if(child.tagName() == "writeable")
			{
				if(child.text() == "true") writeable = true;
				else writeable = false;
			}
			child = child.nextSibling().toElement();
		}
		if(uri.startsWith("ggzmeta://"))
		{
			for(it = m_metaservers.begin(); it != m_metaservers.end(); it++)
				//if((*it) == uri) uri = QString::null;
				if((*it) == uri) it = m_metaservers.remove(it);
			if(!uri.isNull())
			{
				m_metaservers << uri;
				emit signalNewMetaserver(uri);

				config->setGroup("meta");
				config->writeEntry("metaservers", m_metaservers);
				config->setGroup("identifiers");
				cleanurl = uri;
				config->writeEntry(cleanurl.replace("=", "%3d"), name);
				config->setGroup("contributions");
				config->writeEntry(uri, writeable);
				config->sync();
			}
			has_metaserver = true;
		}
		else
		{
			if(!has_stations)
			{
				/*view->clear();*/
				has_stations = true;
			}

			consideration = true;
			qtmp = view->firstChild();

			if(grouping == "bandwidth")
			{
				qtmp2 = view->findItem(speed, 0);
				if(qtmp2) qtmp = qtmp2->firstChild();
			}
			if(grouping == "style")
			{
				qtmp2 = view->findItem(style, 0);
				if(qtmp2) qtmp = qtmp2->firstChild();
			}

			while(qtmp)
			{
				if((qtmp->text(0) == stream) && (qtmp->text(1) == speed) && (qtmp->text(2) == style))
					if((qtmp->text(3) == location) && (qtmp->text(4) == uri) && (qtmp->text(5)) == type)
						consideration = false;
				qtmp = qtmp->nextSibling();
			}

			if(consideration)
			{
				if(grouping == "bandwidth")
				{
					qtmp = view->findItem(speed, 0);
					if(!qtmp) qtmp = new KListViewItem(view, speed);
					tmp = new KListViewItem(qtmp, stream, speed, style, location, uri, type);
				}
				else if(grouping == "style")
				{
					qtmp = view->findItem(style, 0);
					if(!qtmp) qtmp = new KListViewItem(view, style);
					tmp = new KListViewItem(qtmp, stream, speed, style, location, uri, type);

				}
				else
					tmp = new KListViewItem(view, stream, speed, style, location, uri, type);
				tmp->setPixmap(0, QPixmap(d.findResource("data", "kderadiostation/kderadio16.png")));
			}
		}
		node = node.nextSibling();
	}

	if(has_metaserver)
	{
		emit signalNewMetaserver(QString::null);
	}
	if(has_stations)
	{
		emit signalStations();
		if(cache == "yes")
			savecache();
	}

	kdDebug() << "process is ready" << endl;
}

void StationList::slotRead()
{
	QString rdata, rtmp;
	QCString cs;

	while(sock->bytesAvailable())
	{
		kdDebug() << "bytesAvailable " << sock->bytesAvailable() << endl;
/*		rtmp += sock->readLine();*/
		cs.resize(sock->bytesAvailable());
		sock->readBlock(cs.data(), sock->bytesAvailable());
		rtmp += cs;
		sock->waitForMore(500);
	}
	rdata = QString::fromUtf8(rtmp);
	//kdDebug() << "got rawdata(" << rdata << ")" << endl;

	rdata.truncate(rdata.length() - 1);

	process(rdata);

	m_synchronized = 1;

	sock->close();
	//delete sock;
}

void StationList::slotActivate(QListViewItem *item)
{
	bool success;
	bool cache = true;
	QString tmp, normalized;
	KProcess *proc;
	KConfig *config = kapp->config();
	QString output;
	QTimer timer;

	if(item->text(1).isNull()) return;

	success = true;
	if(item->text(5) == "playlist")
	{
		if(cache)
		{
			//KMD5 md5(item->text(4));
			//normalized = QString("kderadiostation-") + QString(md5.hexDigest());
			normalized = QString("kderadiostation-") + item->text(4);
			normalized.replace("/", "_");
			tmp = locateLocal("tmp", normalized);
		}
		else
		{
			KTempFile t(QString::null, ".pls");
			tmp = t.name();
		}
		if(!cache || !QFile::exists(tmp))
		{
			if(KIO::NetAccess::download(item->text(4), tmp NETACCESS_HACK))
				success = true;
			else
			{
				KMessageBox::error(this, i18n("The playlist could not be downloaded."), i18n("Failure"));
				success = false;
			}
		}
	}
	else if(item->text(5) == "stream")
	{
		tmp = item->text(4).latin1();
		success = true;
	}
	else
	{
		KMessageBox::error(this, i18n("Unknown type of audio stream."), i18n("Failure"));
		success = false;
	}

	if(success)
	{
		config->setGroup("settings");
		output = config->readEntry("output", "xmmsspec");

		KStandardDirs d;
		QString path = d.findResource("data", "kderadiostation/specs/");
		KSimpleConfig conf(path + "/" + output);
		conf.setGroup("desktopspec");
		QString exec = conf.readEntry("exec");
		conf.setGroup("techspec");
		QString interface = conf.readEntry("interfaces");

		if(interface == "kmediaplayer")
		{
			bool found;
			QString app;
			DCOPClient *client = kapp->dcopClient();
			client->attach();

			QCStringList list = client->registeredApplications();
			found = false;
			for(QCStringList::iterator it = list.begin(); it != list.end(); it++)
			{
				QString s((*it));
				if(s.startsWith(QString("%1-").arg(exec))) found = 1;
			}

			if(!found)
			{
				proc = new KProcess();
				*proc << exec.latin1();
				proc->start();

				m_synchronized = 0;
				connect(&timer, SIGNAL(timeout()), SLOT(slotTimeout()));
				timer.start(5000, true);

				app = QString("%1-%2").arg(exec).arg(proc->pid());
				while((!client->isApplicationRegistered(app.latin1())) && (!m_synchronized))
					kapp->processEvents();
				timer.stop();
			}

			QByteArray data;
			QDataStream ds(data, IO_WriteOnly);
			ds << KURL(tmp);
			found = client->send(QString("%1-*").arg(exec).latin1(),
				"KMediaPlayer", "openURL(KURL)", data);
			if(!found) kdDebug() << "broadcast for " << exec << "failed" << endl;
		}
		else
		{
			proc = new KProcess();
			*proc << exec.latin1() << tmp.latin1();
			proc->start();
		}
	}
}

void StationList::slotTimeout()
{
	kdDebug() << "timeout!" << endl;

	m_synchronized = 1;
	sock->close();
	//delete sock;
}

void StationList::savecache()
{
	QListViewItem *qtmp;
	QString file;

	file = locateLocal("data", "kderadiostation/cache");
	QFile f(file);
	if(!f.open(IO_WriteOnly)) return;
	QTextStream stream(&f);

	qtmp = view->firstChild();
	while(qtmp)
	{
		while(qtmp->firstChild())
		{
			for(int i = 0; i < qtmp->depth(); i++)
				stream << " ";
			stream << qtmp->text(0) << "\n";
			qtmp = qtmp->firstChild();
		}
		//kdDebug() << "<|||> " << qtmp->text(0) << endl;
		for(int i = 0; i < qtmp->depth(); i++)
			stream << " ";
		for(int i = 0; !qtmp->text(i).isNull(); i++)
			stream << qtmp->text(i) << ":::";
		stream << "\n";
		if(qtmp->nextSibling()) qtmp = qtmp->nextSibling();
		else
		{
			qtmp = qtmp->parent();
			if(qtmp) qtmp = qtmp->nextSibling();
		}
	}

	f.close();
}

void StationList::loadcache()
{
	QListViewItem *qtmp, *qtmp2;
	QString file, line;
	QStringList list;
	int i, depth;
	KStandardDirs d;

	file = locateLocal("data", "kderadiostation/cache");
	QFile f(file);
	if(!f.open(IO_ReadOnly)) return;
	QTextStream stream(&f);

	qtmp = NULL;

	while(!stream.eof())
	{
		line = stream.readLine();
		depth = 0;
		while(line.startsWith(" "))
		{
			line = line.right(line.length() - 1);
			depth++;
		}
		list = QStringList::split(":::", line);

		while((qtmp) && (depth > qtmp->depth() + 1))
			qtmp = qtmp->firstChild();
		while((qtmp) && (depth < qtmp->depth() + 1))
			qtmp = qtmp->parent();
		/*if((!qtmp) && (depth > 0))
		{
			view->setRootIsDecorated(true);
			kdDebug() << "yay for " << line << endl;
			qtmp = view->firstChild();
			kdDebug() << "depth: " << qtmp->depth() << endl;
		}*/

		if(qtmp)
		{
			qtmp2 = new KListViewItem(qtmp);
			view->setRootIsDecorated(true);
		}
		else
		{
			qtmp2 = new KListViewItem(view);
			qtmp = qtmp2;
			qtmp2->setPixmap(0, QPixmap(d.findResource("data", "kderadiostation/kderadio16.png")));
		}
		i = 0;
		for(QStringList::iterator it = list.begin(); it != list.end(); it++, i++)
			qtmp2->setText(i, (*it));
	}

	f.close();
}

