/*
 * KDE Radio Station
 * Copyright (C) 2003 Josef Spillner <spillner@kde.org>
 * Published under GNU GPL conditions.
 */

#ifndef STATIONLIST_H
#define STATIONLIST_H

#include <klistview.h>
#include <qsocket.h>
#include <qstringlist.h>

class QListViewItem;

class StationList : public KListView
{
	Q_OBJECT
	public:
		StationList(QWidget *parent = NULL, const char *name = NULL);
		~StationList();

		QStringList metaservers(bool writeable);
		void addStation(QString metaserver, QString stream, QString uri,
			QString location, QString speed, QString style, QString type);

	signals:
		void signalNewMetaserver(QString uri);
		void signalStations();

	public slots:
		void slotActivate(QListViewItem *item);
		void slotConnected();
		void slotRead();
		void slotError(int error);
		void slotTimeout();

		void slotUpdateMetaservers();
		void slotUpdateStations();

	private:
		void doconnection(QString query);
		void doupdate(QString query, QString uri);
		void process(QString content);
		void processlocal(QString content);
		void processicecast(QString content);

		void savecache();
		void loadcache();

		void startDrag();

		QString m_host;
		int m_port;
		KListView *view;
		QSocket *sock;
		QString m_query;
		QString m_metaquery;
		QString m_curquery;
		QString m_update;
		QStringList m_metaservers;
		int m_synchronized;
};

#endif

